# coding=utf-8
"""
Manages PDF files metadata operations
"""

import hashlib
from pathlib import Path

import pdfrw
from pdfrw.objects import pdfstring

from edlm import __version__
from edlm.convert import Context


def _hash_folder(folder: Path):
    for item in folder.iterdir():
        assert isinstance(item, Path)
        if item.is_file() and item.suffix in ['.md', '.yml', '.tex']:
            yield item.read_bytes()
        elif item.is_dir():
            yield from _hash_folder(item)


def _iterate_over_data(ctx: Context):
    yield from _hash_folder(ctx.source_folder)
    for media_folder in ctx.media_folders:
        for file in media_folder.iterdir():
            yield file.read_bytes()
    for include in ctx.includes:
        if include.is_dir():
            yield from _hash_folder(include)


def _get_document_hash(ctx: Context) -> str:
    hash_ = hashlib.sha1()
    for data in _iterate_over_data(ctx):
        hash_.update(data)
    return hash_.hexdigest()


def skip_file(ctx: Context) -> bool:
    """
    Checks if a file should be skipped

    Tests for:
        - EDLM version
        - index.md
        - template.tex
        - media folders content

    Args:
        ctx: Context

    Returns: True if file should be skipped

    """
    if ctx.out_file.exists():
        pdf = pdfrw.PdfReader(str(ctx.out_file.absolute()))
        creator = pdfstring.PdfString.to_unicode(pdf.Info.Creator)
        if creator != 'EDLM ' + __version__:
            ctx.info('document generated with an older version of EDLM, regenerating')
            return False
        producer = pdfstring.PdfString.to_unicode(pdf.Info.Producer)
        if producer != 'EDLM ' + _get_document_hash(ctx):
            ctx.info('document updated, regenerating')
            return False
        ctx.info('this document has not been modified, skipping it')
        if ctx.regen:
            ctx.info('forcing re-generation of all documents anyway')
            return False

        return True

    return False


def add_metadata_to_pdf(ctx: Context):
    """
    Adds metadata about EDLM version and source files after
    PDF create

    Args:
        ctx: Context
    """
    out_file = str(ctx.out_file.absolute())
    trailer = pdfrw.PdfReader(out_file)
    trailer.Info.Creator = 'EDLM ' + __version__
    trailer.Info.Producer = 'EDLM ' + _get_document_hash(ctx)
    pdfrw.PdfFileWriter(out_file, trailer=trailer).write()



# coding=utf-8
"""
Pandoc external tool
"""

from pathlib import Path

from edlm import HERE
from edlm.external_tools.base import BaseExternalTool


class Pandoc(BaseExternalTool):
    """
    Pandoc external tool
    """
    url = r'https://www.dropbox.com/s/f6pccole9mdkuex/pandoc-2.0.3-windows.zip?dl=1'
    # noinspection SpellCheckingInspection
    hash = '64685f754a28e5d0cfdaf7214e202e53'
    default_archive = Path(HERE, 'pandoc.7z')
    default_install = Path(HERE, 'pandoc')
    expected_version = '2.0.3'

    def get_exe(self) -> Path:  # pragma: no cover
        """
        Returns: Pandoc executable
        """
        if self._exe is None:
            self._exe = Path(Path(self.install_dir), 'pandoc-2.0.3-windows/pandoc.exe').absolute()
        return self._exe

    def get_version(self) -> str:
        """
        Returns: Pandoc version
        """
        if self._version is None:
            self._version = self('--version').split('\n')[0].split(' ')[1]
        return self._version



# coding=utf-8
"""
EDLM external tools
"""

# noinspection PyProtectedMember
from .miktext._miktext import MikTex
# noinspection PyProtectedMember
from .pandoc._pandoc import Pandoc

MIKTEX = MikTex()
PANDOC = Pandoc()



# coding=utf-8
"""
Etcher's Document Library Manager
"""
from pathlib import Path
import warnings

import elib
from pkg_resources import DistributionNotFound, get_distribution

LOGGER = elib.custom_logging.get_logger('EDLM', log_to_file=True)
elib.custom_logging.activate_elib_logging()

HERE = Path(__file__).parent.parent.absolute()

try:
    __version__ = get_distribution('edlm').version
except DistributionNotFound:  # pragma: no cover
    try:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            from setuptools_scm import get_version
            __version__ = get_version()
    except (ImportError, ModuleNotFoundError):
        __version__ = 'not installed'



# coding=utf-8
from pathlib import Path

from edlm.convert._preprocessor._markdown._include import Context, process_includes


def test_no_include():
    ctx = Context()
    ctx.markdown_text = 'some text'
    ctx.includes = []
    process_includes(ctx)
    assert ctx.markdown_text == 'some text'


def test_local_include():
    ctx = Context()
    ctx.source_folder = Path('.').absolute()
    ctx.markdown_text = 'some text\n\n//include "include.md"'
    include = Path('include.md').absolute()
    include.write_text('some other text')
    ctx.includes = [include]
    process_includes(ctx)
    assert ctx.markdown_text == 'some text\n\nsome other text'


def test_unprocessed_include(caplog):
    ctx = Context()
    ctx.source_folder = Path('.').absolute()
    ctx.markdown_text = 'some text\n\n//include "include.md"\n\n//include "nope.md"'
    include = Path('include.md').absolute()
    include.write_text('some other text')
    ctx.includes = [include]
    process_includes(ctx)
    assert ctx.markdown_text == 'some text\n\nsome other text\n\n//include "nope.md"'
    assert 'there are unprocessed "//include" directives:' in caplog.text
    assert ctx.unprocessed_includes == ['line 00005: //include "nope.md"']



# coding=utf-8

from pathlib import Path

import elib
import pytest
from hypothesis import given
from hypothesis import strategies as st
from mockito import mock, verify, when

import edlm.convert._preprocessor._markdown._images as image
from edlm.convert import Context
from edlm.convert._check_for_unused_images import check_for_unused_images


@pytest.fixture(name='media')
def _dummy_media_folder():
    media = Path('./media')
    media.mkdir()
    test_files = [Path(f'./media/{elib.custom_random.random_string()}') for _ in range(10)]
    for file in test_files:
        file.touch()
    yield media


def test_get_image_full_path():
    test_file = Path('./test.png')
    test_file.touch()
    ctx = Context()
    ctx.media_folders = [Path('.')]
    ctx.image_current = test_file.name
    assert image._get_image_full_path(ctx) == str(test_file.absolute())


def test_image_doesnt_exist():
    ctx = Context()
    ctx.media_folders = [Path('.')]
    ctx.image_current = Path('./nope').name
    with pytest.raises(FileNotFoundError):
        image._get_image_full_path(ctx)


@given(width_str=st.from_regex(r'^[\d]{1,5}(cm|mm)$'))
def test_get_correct_width(width_str):
    ctx = Context()
    ctx.image_width_str = width_str
    ctx.image_max_width = 300
    image._get_correct_width(ctx)
    assert ctx.image_width <= 300


def test_get_correct_width_wrong_unit():
    ctx = Context()
    ctx.image_width_str = '9pm'
    ctx.image_max_width = 300
    with pytest.raises(ValueError):
        image._get_correct_width(ctx)


def test_shrink():
    ctx = Context()
    ctx.image_width_str = '50cm'
    ctx.image_max_width = 20
    image._get_correct_width(ctx)
    assert ctx.image_width == 20


def test_get_correct_width_wrong_raw_str():
    ctx = Context()
    ctx.image_width_str = 'x9pm'
    with pytest.raises(ValueError) as exc:
        image._get_correct_width(ctx)
        assert 'x9pm' in exc


def test_process_image_width_no_width_given():
    ctx = Context()
    ctx.image_max_width = 100
    ctx.image_extras = ''
    image._process_image_width(ctx)
    assert ctx.image_extras == '{width="100mm"}'


def test_process_image_width():
    ctx = Context()
    ctx.image_max_width = 200
    ctx.image_extras = '{width="10cm"}'
    image._process_image_width(ctx)
    assert ctx.image_extras == '{width="100mm"}'


def test_check_for_unused_images(caplog, media: Path):
    caplog.set_level(30, 'EDLM')
    caplog.clear()
    ctx = Context()
    ctx.media_folders = [str(media.absolute()), '.']
    ctx.images_used = set()
    check_for_unused_images(ctx)
    for file in media.iterdir():
        assert file.name in caplog.text


def test_check_for_used_images(caplog, media: Path):
    caplog.set_level(30, 'EDLM')
    caplog.clear()
    ctx = Context()
    ctx.media_folders = [str(media.absolute()), '.']
    ctx.images_used = set()
    caplog.clear()
    ctx.images_used = set()
    for file in list(media.iterdir())[5:]:
        ctx.images_used.add(file.name)
    check_for_unused_images(ctx)
    for file in list(media.iterdir())[:5]:
        assert file.name in caplog.text
    for file in list(media.iterdir())[5:]:
        assert file.name not in caplog.text


def test_check_for_no_unused_images(caplog, media: Path):
    caplog.set_level(30, 'EDLM')
    caplog.clear()
    ctx = Context()
    ctx.media_folders = [str(media.absolute()), '.']
    ctx.images_used = set()
    caplog.clear()
    ctx.images_used = set()
    for file in list(media.iterdir()):
        ctx.images_used.add(file.name)
    check_for_unused_images(ctx)
    for file in list(media.iterdir()):
        assert file.name not in caplog.text


def test_check_for_used_images_only_one_media_folder(caplog):
    caplog.set_level(30, 'EDLM')
    caplog.clear()
    ctx = Context()
    ctx.media_folders = ['.']
    ctx.images_used = set()
    check_for_unused_images(ctx)
    assert not caplog.text


def test_process_image():
    ctx = Context()
    ctx.images_used = set()
    match = mock()
    when(match).group('caption').thenReturn('caption')
    when(match).group('picture').thenReturn('.')
    when(match).group('extras').thenReturn('extras')
    when(image)._process_image_width(ctx)
    when(image)._get_image_full_path(ctx).thenReturn('image_current')

    result = image._process_image(ctx, match)
    assert result == '![caption](image_current)extras'

    assert ctx.image_current == 'image_current'
    verify(image)._process_image_width(ctx)
    when(image)._get_image_full_path(ctx)


def test_process_image_http():
    ctx = Context()
    ctx.images_used = set()
    match = mock()
    when(match).group('caption').thenReturn('caption')
    when(match).group('picture').thenReturn('http://something')
    when(match).group('extras').thenReturn('extras')
    when(image)._process_image_width(ctx)
    when(image)._get_image_full_path(ctx).thenReturn('image_current')

    result = image._process_image(ctx, match)
    assert result == '![caption](http://something)extras'

    assert ctx.image_current == 'http://something'
    verify(image)._process_image_width(ctx)
    when(image)._get_image_full_path(ctx)


def test_process_images(media):
    ctx = Context()
    ctx.image_max_width = 200
    ctx.media_folders = [str(media.absolute()), '.']
    pictures = list(media.iterdir())
    ctx.markdown_text = f"""
![picture1](http://picture1.com)
![picture2]({pictures[0].name})
    ![picture3](http://picture1.com)
    """
    image.process_images(ctx)
    print(ctx.markdown_text)
    assert ctx.markdown_text == f"""
![picture1](http://picture1.com){{width="200mm"}}
![picture2]({pictures[0].absolute()}){{width="200mm"}}
    ![picture3](http://picture1.com){{width="200mm"}}
    """



# coding=utf-8
"""
Processes pictures in Markdown text
"""
import re
import typing
from pathlib import Path

from . import Context

RE_PICTURE_LINE = re.compile(r'!\['
                             r'(?P<caption>.*)'
                             r'\]'
                             r'\('
                             r'(?P<picture>.+)'
                             r'\)'
                             r'(?P<extras>.*)')

RE_WIDTH = re.compile(r'{width="(?P<width>.*)"}')
RE_WIDTH_RAW = re.compile(r'(?P<width>[\d]+)(?P<unit>.*)')


def _get_image_full_path(ctx: Context) -> str:
    image_name = Path(ctx.image_current).name
    for folder in ctx.media_folders:
        image_path = Path(folder, image_name)
        if image_path.exists():
            return str(image_path.absolute())

    raise FileNotFoundError(f'picture "{ctx.image_current}" not found in any media folders: {ctx.media_folders}')


def _get_correct_width(ctx: Context):
    match = RE_WIDTH_RAW.match(ctx.image_width_str)
    if not match:
        raise ValueError(ctx.image_width_str)

    width = int(match.group('width'))
    unit = match.group('unit')

    if unit == 'mm':
        pass
    elif unit == 'cm':
        width *= 10
    else:
        raise ValueError(unit)

    if width > ctx.image_max_width:
        ctx.debug(f'adapting width for "{ctx.image_current}" to {ctx.image_max_width}mm')
        ctx.image_width = ctx.image_max_width
    else:
        ctx.image_width = width


def _process_image_width(ctx: Context):
    width_match = RE_WIDTH.match(ctx.image_extras)

    if width_match:
        ctx.image_width_str = width_match.group('width')
        _get_correct_width(ctx)
        ctx.image_extras = RE_WIDTH.sub(f'{{width="{ctx.image_width}mm"}}', ctx.image_extras)
    else:
        ctx.image_width = ctx.image_max_width

    ctx.debug(f'setting default width for picture {ctx.image_current}')
    ctx.image_extras = f'{{width="{ctx.image_width}mm"}}'


def _process_image(ctx: Context, match) -> typing.Optional[str]:
    ctx.image_caption = match.group('caption')
    ctx.image_current = match.group('picture')
    ctx.image_extras = match.group('extras')

    _process_image_width(ctx)

    if not ctx.image_current.startswith('http'):
        ctx.debug(f'processing picture: {ctx.image_current}')
        ctx.images_used.add(Path(ctx.image_current).name)
        ctx.image_current = _get_image_full_path(ctx)

    return f'![{ctx.image_caption}]({ctx.image_current}){ctx.image_extras}'.replace('\\', '\\\\')


def process_images(ctx: Context):
    """
    Processes pictures in Markdown text

    Args:
        ctx: Context
    """
    ctx.info('processing pictures')
    ctx.images_used = set()
    output = []
    for line in ctx.markdown_text.split('\n'):
        match = RE_PICTURE_LINE.search(line)
        if match:
            output.append(RE_PICTURE_LINE.sub(_process_image(ctx, match) or line, line))
        else:
            output.append(line)

    # check_for_unused_images(ctx)

    ctx.debug('processing of pictures finished')
    ctx.markdown_text = '\n'.join(output)



# coding=utf-8
"""
Makes a PDF document from a source folder
"""

import shutil
import urllib.parse
from pathlib import Path

import elib

from edlm import LOGGER
from edlm.convert import Context
from edlm.external_tools import PANDOC

from ._check_for_unused_images import check_for_unused_images
from ._get_includes import get_includes
from ._get_index import get_index_file
from ._get_media_folders import get_media_folders
from ._get_settings import get_settings
from ._get_template import get_template
from ._pdf_info import add_metadata_to_pdf, skip_file
from ._preprocessor import process_markdown, process_tex_template
from ._temp_folder import TempDir

WIDTH_MODIFIER = 0.8

PAPER_FORMATS_WIDTH = {
    'a0': 841 * WIDTH_MODIFIER,
    'a1': 594 * WIDTH_MODIFIER,
    'a2': 420 * WIDTH_MODIFIER,
    'a3': 29 * WIDTH_MODIFIER,
    'a4': 210 * WIDTH_MODIFIER,
    'a5': 148 * WIDTH_MODIFIER,
    'a6': 105 * WIDTH_MODIFIER,
    'a7': 74 * WIDTH_MODIFIER,
}

BASE_URL = r'http://132virtualwing.org/docs/'


def _download_existing_file(ctx: Context):
    if not ctx.out_file.exists():
        ctx.info(f'trying to download {ctx.out_file.name}')
        url = BASE_URL + urllib.parse.quote(ctx.out_file.name)
        if elib.downloader.download(url, ctx.out_file):
            ctx.info('download successful')
        else:
            ctx.info('download failed')


def _set_max_image_width(ctx: Context):
    paper_size = ctx.paper_size.lower()
    if paper_size not in PAPER_FORMATS_WIDTH:
        raise ValueError(paper_size)
    ctx.image_max_width = int(PAPER_FORMATS_WIDTH[paper_size])


def _remove_artifacts():
    for item in Path('.').iterdir():
        assert isinstance(item, Path)
        if item.is_dir() and item.name.startswith('__TMP') or item.name.startswith('tex2pdf.'):
            shutil.rmtree(str(item.absolute()))


def _build_folder(ctx: Context):
    ctx.info(f'making PDF')

    with TempDir(ctx):

        get_media_folders(ctx)

        get_template(ctx)

        get_index_file(ctx)

        get_settings(ctx)

        get_includes(ctx)

        ctx.template_file = Path(ctx.temp_dir, 'template.tex').absolute()

        title = ctx.source_folder.name
        ctx.title = title

        out_folder = elib.path.ensure_dir('.', must_exist=False, create=True)
        ctx.out_folder = out_folder

        for paper_size in ctx.settings.papersize:
            ctx.paper_size = paper_size
            _set_max_image_width(ctx)

            if paper_size.lower() == 'a4' or len(ctx.settings.papersize) == 1:
                ctx.out_file = Path(out_folder, f'{title}.PDF').absolute()
            else:
                ctx.out_file = Path(out_folder, f'{title}_{paper_size}.PDF').absolute()

            _download_existing_file(ctx)

            if skip_file(ctx):
                continue

            process_markdown(ctx)

            check_for_unused_images(ctx)

            process_tex_template(ctx)

            ctx.source_file = Path(ctx.temp_dir, 'source.md').absolute()
            ctx.source_file.write_text(ctx.markdown_text, encoding='utf8')

            ctx.info(f'building format: {paper_size}')

            ctx.debug(f'context:\n{elib.pretty_format(ctx.__repr__())}')

            # noinspection SpellCheckingInspection
            pandoc_cmd = [
                '-s',
                '--toc',
                f'--template "{ctx.template_file}"',
                f'--listings "{ctx.source_file}"',
                f'-o "{ctx.out_file}"',
                '-V geometry:margin=1.5cm',
                '-V test',
                '-V geometry:headheight=17pt',
                '-V geometry:includehead',
                '-V geometry:includefoot',
                '-V geometry:heightrounded',
                '-V lot',
                '-V lof',
                '--pdf-engine=xelatex',
                f'-V papersize:{ctx.paper_size}',
                '-N',
            ]

            PANDOC(' '.join(pandoc_cmd))

            add_metadata_to_pdf(ctx)


def _is_source_folder(folder: Path) -> bool:
    if not folder.is_dir():
        return False
    return Path(folder, 'index.md').exists()


def make_pdf(ctx: Context, source_folder: Path):
    """
    Makes a PDF document from a source folder

    Args:
        ctx: Context
        source_folder: source folder
    """
    _remove_artifacts()

    source_folder = elib.path.ensure_dir(source_folder).absolute()

    LOGGER.info(f'analyzing folder: "{source_folder}"')

    if _is_source_folder(source_folder):
        ctx.source_folder = source_folder
        _build_folder(ctx)

    else:
        for child in source_folder.iterdir():
            if _is_source_folder(child):
                ctx.source_folder = child.absolute()
                _build_folder(ctx)



# coding=utf-8
"""
Processes aliases
"""

from . import Context


def process_aliases(ctx: Context):
    """
    Processes aliases
    """
    ctx.debug('processing aliases')
    for alias, value in ctx.settings.aliases.items():
        ctx.debug(f'processing alias: {alias}')
        ctx.markdown_text = ctx.markdown_text.replace(alias, value)



# coding=utf-8

from pathlib import Path

import pytest

from edlm.convert import Context, ConvertError
from edlm.convert import _get_template as template


def test_no_template():
    ctx = Context()
    ctx.source_folder = Path('.').absolute()
    with pytest.raises(ConvertError):
        template.get_template(ctx)


def test_simple_template():
    ctx = Context()
    ctx.source_folder = Path('.').absolute()
    template_file = Path('./template.tex').absolute()
    template_file.touch()
    template.get_template(ctx)
    assert ctx.template_source == template_file


def test_multiple_template():
    ctx = Context()
    ctx.source_folder = Path('.').absolute()

    sub1 = Path('./sub1').absolute()
    sub1.mkdir()
    template_file1 = Path(sub1, 'template.tex').absolute()
    template_file1.touch()
    sub2 = Path(sub1, 'sub2').absolute()
    sub2.mkdir()
    template_file2 = Path(sub2, 'template.tex').absolute()
    template_file2.touch()
    sub3 = Path(sub2, 'sub3').absolute()
    sub3.mkdir()
    template_file3 = Path(sub3, 'template.tex').absolute()
    template_file3.touch()

    ctx.source_folder = Path(sub3).absolute()

    template.get_template(ctx)
    assert ctx.template_source == template_file3

    template_file3.unlink()

    template.get_template(ctx)
    assert ctx.template_source == template_file2

    template_file2.unlink()

    template.get_template(ctx)
    assert ctx.template_source == template_file1



# coding=utf-8
"""
Processing of Markdown content
"""

from ..._context import Context
from ._aliases import process_aliases
from ._images import process_images
from ._references import process_references
from ._include import process_includes


def process_markdown(ctx: Context):
    """
    Processing of Markdown content
    """
    ctx.debug(f'processing markdown file: {ctx.index_file}')
    process_aliases(ctx)
    process_images(ctx)
    process_references(ctx)
    process_includes(ctx)



# coding=utf-8
"""
Base class for external tools
"""
import functools
import os
import sys
import typing
from pathlib import Path

import elib
import pyunpack

from edlm import LOGGER

LOGGER = LOGGER

STR_OR_PATH = typing.Union[str, Path]


@functools.lru_cache(1)
def _find_patool(path=sys.executable) -> Path:
    LOGGER.debug('looking for Patool')
    python_exe = Path(path).absolute()
    LOGGER.debug(f'python_exe: {python_exe}')
    patool_path = Path(python_exe.parent, 'patool')
    LOGGER.debug(f'looking at {patool_path}')
    if patool_path.exists():
        LOGGER.debug(f'patool found: {patool_path}')
        return patool_path.absolute()

    scripts_path = Path(python_exe.parent, 'scripts')
    if scripts_path.exists() and scripts_path.is_dir():
        patool_path = Path(scripts_path, 'patool')
        LOGGER.debug(f'looking at {patool_path}')
        if patool_path.exists():
            LOGGER.debug(f'patool found: {patool_path}')
            return patool_path.absolute()

    raise FileNotFoundError('patool not found')


class BaseExternalTool:
    """
    Base class for external tools
    """
    url = None
    hash = None
    default_archive = None
    default_install = None
    expected_version = None

    def __init__(self, archive: STR_OR_PATH = None, install_dir: STR_OR_PATH = None):
        self.__check_values()

        if archive is None:
            archive = self.default_archive
        if install_dir is None:
            install_dir = self.default_install

        self.archive = elib.path.ensure_file(archive, must_exist=False).absolute()
        self.install_dir = elib.path.ensure_dir(install_dir, must_exist=False).absolute()

        LOGGER.debug(f'{self.__class__.__name__}: archive: {self.archive}')
        LOGGER.debug(f'{self.__class__.__name__}: install dir: {self.install_dir}')

        self._exe = None
        self._version = None

    def __call__(self, cmd: str):
        out, _ = elib.run(str(self.get_exe().absolute()) + ' ' + cmd, mute=True)
        return out

    def __check_values(self):
        for value, name in (
                (self.default_install, 'default_install'),
                (self.default_archive, 'default_archive'),
                (self.url, 'url'),
                (self.hash, 'hash'),
                (self.expected_version, 'expected_version'),
        ):
            if value is None:
                raise ValueError(
                    f'Class "{self.__class__.__name__}(BaseExternalTool)" is missing default value for "{name}"')

    def _archive_exists(self) -> bool:  # pragma: no cover
        return self.archive.exists()

    def _archive_is_correct(self) -> bool:
        if not self._archive_exists():
            return False
        return elib.hash_.get_hash(self.archive.read_bytes()) == self.hash

    def _is_installed(self) -> bool:
        LOGGER.debug(f'{self.__class__.__name__}: checking installation')
        if not self.get_exe().exists():
            LOGGER.debug(f'{self.__class__.__name__}: executable not found')
            return False
        if not self.get_version() == self.expected_version:
            LOGGER.debug(f'{self.__class__.__name__}: wrong version: '
                         f'{self.get_version()} '
                         f'(expected {self.expected_version})')
            return False
        return True

    # pylint: disable=inconsistent-return-statements
    def _download(self) -> bool:
        if self._archive_is_correct():
            LOGGER.debug('{self.__class__.__name__}: archive already downloaded')
            return True

        LOGGER.debug(f'{self.__class__.__name__}: downloading: {self.url} -> {self.archive}')
        if elib.downloader.download(
                url=self.url,
                outfile=self.archive.absolute(),
                hexdigest=self.hash
        ):
            LOGGER.debug(f'{self.__class__.__name__}: download successful')
            return True
        else:
            LOGGER.error(f'{self.__class__.__name__}: download failed')
            exit(-1)

    def _extract(self):
        LOGGER.info(f'{self.__class__.__name__}: extracting (this may take a while)')
        archive = pyunpack.Archive(self.archive)
        patool = _find_patool()
        archive.extractall(directory=self.install_dir, auto_create_dir=True, patool_path=str(patool))
        LOGGER.debug(f'{self.__class__.__name__}: removing archive')
        # self.archive.unlink()
        LOGGER.info(f'{self.__class__.__name__}: successfully extracted')

    def setup(self):
        """
        Setup this tool
        """
        if not self._is_installed():
            LOGGER.debug(f'{self.__class__.__name__}: setting up')
            self._download()
            self._extract()

        LOGGER.debug(f'{self.__class__.__name__}: adding to PATH: {self.get_exe().parent.absolute()}')
        os.environ['PATH'] = f'{self.get_exe().parent.absolute()};' + os.environ['PATH']
        LOGGER.debug(f'{self.__class__.__name__}: {self.get_version()}')

    def get_exe(self) -> Path:
        """
        Returns: executable for this tool
        """
        raise NotImplementedError(f'missing exe for {self.__class__.__name__}')

    def get_version(self) -> str:
        """
        Returns: version for this tool
        """
        raise NotImplementedError(f'missing exe for {self.__class__.__name__}')



# coding=utf-8
"""
Checks that all images that in this document media folder
are used throughout the document, otherwise issue a warning
"""
from pathlib import Path

import elib


def check_for_unused_images(ctx):
    """
    Checks that all images that in this document media folder
    are used throughout the document, otherwise issue a warning

    Args:
        ctx: Context
    """

    unused_images = set()
    if len(ctx.media_folders) > 1:
        ctx.debug('checking media path for unused pictures')
        for file in Path(ctx.media_folders[0]).iterdir():
            if file.name not in ctx.images_used:
                ctx.debug(f'checking that "{file}" is used')
                unused_images.add(str(file.absolute()))

        if unused_images:
            ctx.warning(f'unused files found:\n{elib.pretty_format(sorted(unused_images))}')



# coding=utf-8
"""
Gathers "index.md" file
"""

import elib

from ._context import Context


def get_index_file(ctx: Context):
    """
    Gathers "index.md" file
    """
    index_file = elib.path.ensure_file(ctx.source_folder, 'index.md')
    ctx.debug(f'index file: {index_file}')
    ctx.index_file = index_file
    ctx.markdown_text = ctx.index_file.read_text(encoding='utf8')



# coding=utf-8
"""
Gathers settings
"""

import collections
from pathlib import Path

import elib
import yaml

from ._context import Context
from ._exc import ConvertError

from._settings import Settings


def update_nested_dict(source_dict, updated_dict):
    """
    Updates a dictionary from another

    Args:
        source_dict: source dictionary (will be overwritten)
        updated_dict: updated dictionary (will take precedence)

    Returns: merged dictionary

    """
    for key, value in updated_dict.items():
        if isinstance(value, collections.Mapping):
            result = update_nested_dict(source_dict.get(key, {}), value)
            source_dict[key] = result
        else:
            source_dict[key] = updated_dict[key]
    return source_dict


def get_settings(ctx: Context):
    """
    Gathers settings
    """
    ctx.info('reading settings')

    ctx.settings = Settings()
    ctx.settings_files = []

    this_folder = ctx.source_folder
    while True:
        ctx.debug(f'traversing: {this_folder}')
        file = Path(this_folder, 'settings.yml')
        if file.exists() and file.is_file():
            ctx.debug(f'settings file found: {file}')
            ctx.settings_files.append(file)
        if len(this_folder.parents) is 1:
            ctx.debug(f'reach mount point at: "{this_folder}"')
            break
        this_folder = this_folder.parent

    if not ctx.settings_files:
        raise ConvertError('no "settings.yml" file found', ctx)

    for file in reversed(ctx.settings_files):
        if not file.read_text(encoding='utf8'):
            raise ConvertError(f'empty "settings.yml": {file.absolute()}', ctx)
        with open(file) as stream:
            these_settings = yaml.load(stream)
            ctx.debug(f'content of "{file}": {elib.pretty_format(these_settings)}')
        ctx.settings.update(these_settings)

    ctx.debug(f'settings files:\n{elib.pretty_format(ctx.settings_files)}')
    ctx.debug(f'settings:\n{elib.pretty_format(ctx.settings)}')



# coding=utf-8

import typing
from pathlib import Path

import elib
import elib._run
import pytest
from mockito import verify, verifyStubbedInvocationsAreUsed, when

from edlm.convert import Context, _make_pdf
from edlm.external_tools import PANDOC


@pytest.mark.parametrize(
    'format_',
    list(_make_pdf.PAPER_FORMATS_WIDTH.keys())
)
def test_max_image_width(format_):
    ctx = Context()
    ctx.paper_size = format_
    _make_pdf._set_max_image_width(ctx)
    assert ctx.image_max_width == int(_make_pdf.PAPER_FORMATS_WIDTH[format_])


@pytest.mark.parametrize(
    'format_',
    ('A8', 'A9', 'B4', 'nil', '1')
)
def test_max_image_wrong_value(format_):
    ctx = Context()
    ctx.paper_size = format_
    with pytest.raises(ValueError):
        _make_pdf._set_max_image_width(ctx)


def test_remove_artifacts():
    for _ in range(30):
        Path(f'./__TMP{elib.custom_random.random_string()}').mkdir()
        Path(f'./tex2pdf.{elib.custom_random.random_string()}').mkdir()
    _make_pdf._remove_artifacts()
    assert not list(Path('.').iterdir())


def _get_standard_build_folder() -> typing.Tuple[Context, Path]:
    ctx = Context()
    pandoc = Path('pandoc')
    pandoc.touch()
    src_folder = Path('./test').absolute()
    src_folder.mkdir()
    ctx.source_folder = src_folder
    source_file = Path('./test/index.md')
    source_file.touch()
    ctx.source_file = source_file
    ctx.markdown_text = ''
    return ctx, pandoc


@pytest.mark.parametrize(
    'paper_size',
    (['a4'], ['a3', 'a5'], ['a3'])
)
def test_build_folder(paper_size):
    ctx, pandoc = _get_standard_build_folder()
    ctx.settings.papersize = paper_size
    ctx.media_folders = []
    when(_make_pdf).get_media_folders(...)
    when(_make_pdf).skip_file(...)
    when(_make_pdf).add_metadata_to_pdf(ctx)
    when(_make_pdf).get_template(...)
    when(_make_pdf).get_index_file(...)
    when(_make_pdf).get_settings(...)
    when(_make_pdf).process_markdown(...)
    when(_make_pdf).process_tex_template(...)

    when(PANDOC).get_exe().thenReturn(pandoc)
    when(elib).run(...).thenReturn(('out', 0))
    _make_pdf._build_folder(ctx)
    verifyStubbedInvocationsAreUsed()
    assert ctx.title == 'test'


def test_build_folder_skip():
    ctx, _ = _get_standard_build_folder()
    ctx.settings.papersize = ['a4']
    when(_make_pdf).get_media_folders(...)
    when(_make_pdf).skip_file(...).thenReturn(True)
    when(_make_pdf).get_template(...)
    when(_make_pdf).get_index_file(...)
    when(_make_pdf).get_settings(...)
    _make_pdf._build_folder(ctx)
    verifyStubbedInvocationsAreUsed()


def test_is_source_folder():
    assert _make_pdf._is_source_folder(Path('.')) is False
    Path('./index.md').touch()
    assert _make_pdf._is_source_folder(Path('.')) is True


def test_is_source_folder_is_file():
    file = Path('./file')
    file.touch()
    assert _make_pdf._is_source_folder(file) is False


def test_make_pdf_empty_folder():
    ctx = Context()
    when(_make_pdf)._build_folder(ctx)
    _make_pdf.make_pdf(ctx, Path('.'))
    verify(_make_pdf, times=0)._build_folder(ctx)


def test_make_pdf_document_folder():
    ctx = Context()
    Path('./index.md').touch()
    when(_make_pdf)._build_folder(ctx)
    _make_pdf.make_pdf(ctx, Path('.'))
    verify(_make_pdf, times=1)._build_folder(ctx)


def test_make_pdf_parent_folder():
    ctx = Context()
    for sub_folder in [
        Path('./sub1'),
        Path('./sub2'),
        Path('./sub3'),
        Path('./sub4'),
    ]:
        sub_folder.mkdir()
        Path(sub_folder, 'index.md').touch()
    Path('./sub5').mkdir()
    when(_make_pdf)._build_folder(ctx)
    _make_pdf.make_pdf(ctx, Path('.'))
    verify(_make_pdf, times=4)._build_folder(ctx)


def test_download_existing_file(caplog):
    file = Path('test').absolute()
    ctx = _make_pdf.Context()
    ctx.out_file = file
    when(elib.downloader).download(...).thenReturn(True)
    _make_pdf._download_existing_file(ctx)
    assert 'download successful' in caplog.text
    when(elib.downloader).download(...).thenReturn(False)
    _make_pdf._download_existing_file(ctx)
    assert 'download failed' in caplog.text


def test_download_existing_file_file_exists(caplog):
    file = Path('test').absolute()
    ctx = _make_pdf.Context()
    ctx.out_file = file
    file.touch()
    _make_pdf._download_existing_file(ctx)
    assert 'download failed' not in caplog.text
    assert 'download successful' not in caplog.text



# coding=utf-8

from pathlib import Path

import elib
from mockito import contains, verifyStubbedInvocationsAreUsed, when

from edlm import external_tools


def test_pandoc():
    pandoc = external_tools.Pandoc()
    exe = Path('exe')
    when(pandoc).get_exe().thenReturn(exe)
    when(elib).run(contains('--version'), mute=True).thenReturn(('pandoc version\ntext', 0))
    assert pandoc.get_version() == 'version'
    assert pandoc.get_version() == 'version'
    verifyStubbedInvocationsAreUsed()



# coding=utf-8

from pathlib import Path

from edlm.convert import Context
from edlm.convert._get_media_folders import get_media_folders

# def test_not_media_folder():
#     ctx = Context()
#     ctx.source_folder = Path('.').absolute()
#     with pytest.raises(ConvertError):
#         get_media_folders(ctx)


def test_get_media_folder():
    ctx = Context()
    ctx.source_folder = Path('.').absolute()
    media = Path('./media').absolute()
    media.mkdir()
    get_media_folders(ctx)
    assert ctx.media_folders == [media]


def test_get_media_folder_multiple():
    sub1 = Path('./sub1').absolute()
    sub1.mkdir()
    media1 = Path(sub1, 'media').absolute()
    media1.mkdir()
    sub2 = Path(sub1, 'sub2').absolute()
    sub2.mkdir()
    media2 = Path(sub2, 'media').absolute()
    media2.mkdir()
    sub3 = Path(sub2, 'sub3').absolute()
    sub3.mkdir()
    media3 = Path(sub3, 'media').absolute()
    media3.mkdir()
    ctx = Context()
    ctx.source_folder = sub3
    get_media_folders(ctx)
    assert ctx.media_folders == [media3, media2, media1]



# coding=utf-8

from pathlib import Path

import elib
import pytest
import pyunpack
from mockito import mock, verify, verifyStubbedInvocationsAreUsed, when

import edlm.external_tools.base
from edlm.external_tools.base import BaseExternalTool, _find_patool


def test_find_patool_not_found():
    with pytest.raises(FileNotFoundError):
        _find_patool('.')


def test_patool_alongside():
    dummy = Path('dummy').absolute()
    patool = Path('patool').absolute()
    patool.touch()
    assert _find_patool(dummy) == patool


def test_patool_in_scripts_folder():
    scripts = Path('./scripts').absolute()
    scripts.mkdir()
    dummy = Path('dummy').absolute()
    patool = Path(scripts, 'patool').absolute()
    patool.touch()
    assert _find_patool(dummy) == patool
    patool.unlink()
    _find_patool.cache_clear()
    with pytest.raises(FileNotFoundError):
        _find_patool(dummy)


class Sub(BaseExternalTool):
    url = 'None'
    hash = 'None'
    default_archive = 'None'
    default_install = 'None'
    expected_version = 'None'

    def get_exe(self):
        return self._exe

    def get_version(self):
        return 'version'


def test_base_empty_init():
    sub = Sub()
    assert sub.url == 'None'
    assert sub.hash == 'None'
    assert sub.archive == Path('None').absolute()
    assert sub.install_dir == Path('None').absolute()
    assert sub.expected_version == 'None'


def test_base_meaningful_init():
    sub = Sub('archive', 'install_dir')
    assert sub.url == 'None'
    assert sub.hash == 'None'
    assert sub.archive == Path('archive').absolute()
    assert sub.install_dir == Path('install_dir').absolute()
    assert sub.expected_version == 'None'


def test_call():
    when(elib).run('test some command', mute=True).thenReturn(('out', None))
    sub = Sub()
    exe = mock()
    when(exe).absolute().thenReturn('test')
    sub._exe = exe
    assert sub('some command') == 'out'
    verifyStubbedInvocationsAreUsed()


@pytest.mark.parametrize(
    'attrib',
    ('default_install', 'default_archive', 'url', 'hash', 'expected_version')
)
def test_check_values(attrib):
    class Dummy(Sub):
        pass

    setattr(Dummy, attrib, None)
    with pytest.raises(ValueError):
        Dummy()


def test_archive_exists():
    sub = Sub()
    assert sub._archive_exists() is False
    archive = Path('./archive')
    archive.touch()
    sub.archive = archive
    assert sub._archive_exists() is True


def test_archive_is_correct():
    sub = Sub()
    assert sub._archive_is_correct() is False
    archive = Path('./archive')
    archive.touch()
    sub.archive = archive
    assert sub._archive_is_correct() is False
    sub.hash = elib.hash_.get_hash(archive.read_bytes())
    assert sub._archive_is_correct() is True


def test_archive_is_installed():
    sub = Sub()
    archive = Path('./archive')
    archive.touch()
    sub.archive = archive
    exe = Path('./exe')
    sub._exe = exe
    sub.archive = archive
    sub.hash = elib.hash_.get_hash(archive.read_bytes())
    assert sub._is_installed() is False
    exe.touch()
    assert sub._is_installed() is False
    sub.expected_version = 'version'
    assert sub._is_installed() is True


def test_download():
    when(elib.downloader).download(...).thenReturn(True)
    sub = Sub()
    assert sub._download() is True
    verify(elib.downloader).download(...)


def test_download_already_correct():
    when(elib.downloader).download(...).thenReturn(True)
    sub = Sub()
    archive = Path('./archive')
    archive.touch()
    sub.archive = archive
    exe = Path('./exe')
    sub._exe = exe
    sub.archive = archive
    sub.hash = elib.hash_.get_hash(archive.read_bytes())
    assert sub._download() is True
    verify(elib.downloader, times=0).download(...)


def test_download_failed():
    when(elib.downloader).download(...).thenReturn(False)
    sub = Sub()
    with pytest.raises(SystemExit):
        sub._download()


def test_extract():
    when(edlm.external_tools.base)._find_patool().thenReturn('patool')
    sub = Sub()
    archive = Path('./archive')
    archive.touch()
    sub.archive = archive
    exe = Path('./exe')
    sub._exe = exe
    sub.archive = archive
    sub.hash = elib.hash_.get_hash(archive.read_bytes())
    when(pyunpack.Archive).extractall(...)
    sub._extract()
    verify(pyunpack.Archive).extractall(...)


def test_setup():
    when(BaseExternalTool)._download()
    when(BaseExternalTool)._extract()
    sub = Sub()
    exe = Path('./exe')
    sub._exe = exe
    sub.setup()
    verify(BaseExternalTool)._download()
    verify(BaseExternalTool)._extract()
    archive = Path('./archive')
    archive.touch()
    sub.archive = archive
    sub.archive = archive
    sub.hash = elib.hash_.get_hash(archive.read_bytes())
    assert sub._is_installed() is False
    exe.touch()
    assert sub._is_installed() is False
    sub.expected_version = 'version'
    sub.setup()
    verify(BaseExternalTool)._download()
    verify(BaseExternalTool)._extract()



# coding=utf-8

from pathlib import Path

import pytest

from edlm.convert._context import Context, _Val


@pytest.mark.parametrize(
    'level',
    ['debug', 'info', 'warning', 'error']
)
def test_logger(caplog, level):
    caplog.clear()
    ctx = Context()
    func = getattr(ctx, level)
    assert 'test' not in caplog.text
    func('test')
    assert 'test' in caplog.text


@pytest.mark.parametrize(
    'level',
    ['debug', 'info', 'warning', 'error']
)
def test_logger_with_index_file(caplog, level):
    caplog.clear()
    ctx = Context()
    index_file = Path('index.md').absolute()
    ctx.index_file = index_file
    func = getattr(ctx, level)
    assert 'test' not in caplog.text
    func('test')
    assert 'test' in caplog.text
    assert str(index_file) in caplog.text


def test_no_default():
    class Dummy(Context):
        no_default = _Val(str)

    dummy = Dummy()
    with pytest.raises(KeyError):
        _ = dummy.no_default


def test_wrong_type():
    ctx = Context()
    with pytest.raises(TypeError):
        ctx.title = 1


def test_sub_context():
    ctx = Context()
    sub_context = ctx.get_sub_context()
    assert isinstance(sub_context, Context)
    assert isinstance(sub_context.data, dict)
    for k, v in sub_context.data.items():
        if k in ['settings']:
            continue
        assert ctx.data[k] == v



# coding=utf-8

from pathlib import Path

import elib
from click.testing import CliRunner
from mockito import ANY, verify, when

import edlm.cli
import edlm.convert


def test_basic(runner: CliRunner):
    result = runner.invoke(edlm.cli.cli)
    assert result.exit_code == 0
    assert 'Usage: cli [OPTIONS] COMMAND [ARGS]...' in result.output


def test_debug(runner: CliRunner):
    when(elib.custom_logging).set_handler_level(...)
    when(edlm.cli.PANDOC).setup(...)
    when(edlm.cli.MIKTEX).setup(...)
    when(edlm.cli).make_pdf(...)
    runner.invoke(edlm.cli.cli, ['convert', 'pdf'])
    verify(edlm.cli.PANDOC).setup(...)
    verify(edlm.cli.MIKTEX).setup(...)
    verify(elib.custom_logging).set_handler_level('EDLM', 'ch', 'info')
    runner.invoke(edlm.cli.cli, ['--debug', 'convert', 'pdf'])
    verify(elib.custom_logging).set_handler_level('EDLM', 'ch', 'debug')


def test_make_pdf(runner: CliRunner):
    folder1 = Path('./folder1').absolute()
    folder2 = Path('./folder2').absolute()
    folder3 = Path('./folder3').absolute()
    folder1.mkdir()
    folder2.mkdir()
    folder3.mkdir()
    when(edlm.cli.PANDOC).setup(...)
    when(edlm.cli.MIKTEX).setup(...)
    when(edlm.cli).make_pdf(...)
    runner.invoke(edlm.cli.cli, ['convert', 'pdf', 'folder1', 'folder2', 'folder3'])
    verify(edlm.cli.PANDOC).setup(...)
    verify(edlm.cli.MIKTEX).setup(...)
    verify(edlm.cli).make_pdf(ANY, str(folder1))
    verify(edlm.cli).make_pdf(ANY, str(folder2))
    verify(edlm.cli).make_pdf(ANY, str(folder3))



# coding=utf-8

from pathlib import Path

from edlm.convert._get_includes import Context, get_includes


def test_no_includes():
    ctx = Context()
    ctx.source_folder = Path('.').absolute()
    ctx.markdown_text = 'some text'
    get_includes(ctx)
    assert not ctx.includes


def test_local_includes():
    ctx = Context()
    ctx.source_folder = Path('.').absolute()
    ctx.markdown_text = 'some text\n\n//include "include.md"'
    include = Path('include.md').absolute()
    include.write_text('plop', encoding='utf8')
    include2 = Path('include2.md').absolute()
    include2.write_text('plop', encoding='utf8')
    get_includes(ctx)
    assert ctx.includes == [include]


def test_external_includes():
    ctx = Context()
    source_folder = Path('./plop').absolute()
    source_folder.mkdir()
    ctx.source_folder = source_folder
    ctx.markdown_text = 'some text\n\n//include "some doc"'
    include_folder = Path('some doc').absolute()
    include_folder.mkdir()
    include = Path(include_folder, 'index.md').absolute()
    include.write_text('plop', encoding='utf8')
    include_folder2 = Path('some other doc').absolute()
    include_folder2.mkdir()
    include2 = Path(include_folder, 'index.md').absolute()
    include2.write_text('plop', encoding='utf8')
    get_includes(ctx)
    assert ctx.includes == [include_folder]



# coding=utf-8

from pathlib import Path

import pdfrw
from mockito import mock, verify, when

from edlm import __version__
from edlm.convert import _pdf_info


def test_get_document_hash():
    ctx = _pdf_info.Context()
    ctx.source_folder = Path('.')
    index_file = Path('index.md')
    index_file.write_text('dsfsdfsdfsdfsdf', encoding='utf8')
    template = Path('template.tex')
    template.write_text('sxdvxfgsfdsdf', encoding='utf8')
    dummy = Path('dummy')
    dummy.touch()
    settings = Path('settings.yml')
    settings.write_text('cxvsdfgqsdfqsdq', encoding='utf8')
    media = Path('./media')
    media.mkdir()
    pic = Path(media, 'pic')
    pic.write_text('sdvsdfsdfqazfqsdf', encoding='utf8')
    ctx.index_file = index_file
    ctx.template_source = template
    ctx.settings_files = [settings]
    ctx.media_folders = [media]
    include_dir = Path('include')
    include_dir.mkdir()
    include_file = Path(include_dir, 'include.md')
    include_file.write_text('sdfsdf', encoding='utf8')
    ctx.includes = [include_dir, include_file]
    hash_ = _pdf_info._get_document_hash(ctx)
    assert hash_ == _pdf_info._get_document_hash(ctx)
    index_file.write_text('sdfsfsaqzzdqscxdcg', encoding='utf8')
    dummy.write_text('sdfsdvcsdvxwdvsdf', encoding='utf8')
    assert hash_ != _pdf_info._get_document_hash(ctx)
    hash_ = _pdf_info._get_document_hash(ctx)
    assert hash_ == _pdf_info._get_document_hash(ctx)
    template.write_text('sdgfsdfqljkqslkjdh', encoding='utf8')
    dummy.write_text('sdfsdvcsdqsdqsdvxwdvsdf', encoding='utf8')
    assert hash_ != _pdf_info._get_document_hash(ctx)
    hash_ = _pdf_info._get_document_hash(ctx)
    assert hash_ == _pdf_info._get_document_hash(ctx)
    settings.write_text('sdgfsdfqljkqslkjdh', encoding='utf8')
    dummy.write_text('qsdqsdqdazazreqsdf', encoding='utf8')
    assert hash_ != _pdf_info._get_document_hash(ctx)
    hash_ = _pdf_info._get_document_hash(ctx)
    assert hash_ == _pdf_info._get_document_hash(ctx)
    pic.write_text('sdgfsdfqljkqslkjdh', encoding='utf8')
    dummy.write_text('qsddsdds', encoding='utf8')
    assert hash_ != _pdf_info._get_document_hash(ctx)
    hash_ = _pdf_info._get_document_hash(ctx)
    assert hash_ == _pdf_info._get_document_hash(ctx)
    include_file.write_text('wxcwxcmlml', encoding='utf8')
    assert hash_ != _pdf_info._get_document_hash(ctx)
    hash_ = _pdf_info._get_document_hash(ctx)
    assert hash_ == _pdf_info._get_document_hash(ctx)


def test_skip_file_doesnt_exist():
    ctx = _pdf_info.Context()
    ctx.out_file = Path('./test')
    assert not _pdf_info.skip_file(ctx)


def test_skip_file_should_skip(caplog):
    ctx = _pdf_info.Context()
    ctx.out_file = Path('./test')
    pdf = mock()
    pdf.Info = mock()
    pdf.Info.Creator = 'creator'
    pdf.Info.Producer = 'producer'
    Path('./test').touch()
    when(pdfrw).PdfReader(...).thenReturn(pdf)
    when(pdfrw.objects.pdfstring.PdfString).to_unicode(...) \
        .thenReturn('EDLM ' + __version__) \
        .thenReturn('EDLM hash')
    when(_pdf_info)._get_document_hash(...).thenReturn('hash')
    assert _pdf_info.skip_file(ctx)
    assert 'this document has not been modified, skipping it' in caplog.text


def _get_should_skip_context(hash_='EDLM hash', version=__version__) -> _pdf_info.Context:
    ctx = _pdf_info.Context()
    ctx.out_file = Path('./test')
    pdf = mock()
    pdf.Info = mock()
    pdf.Info.Creator = 'creator'
    pdf.Info.Producer = 'producer'
    Path('./test').touch()
    when(pdfrw).PdfReader(...).thenReturn(pdf)
    when(pdfrw.objects.pdfstring.PdfString).to_unicode(...) \
        .thenReturn('EDLM ' + version) \
        .thenReturn(hash_)
    when(_pdf_info)._get_document_hash(...).thenReturn('hash')
    return ctx


def test_skip_file_force_regen(caplog):
    ctx = _get_should_skip_context()
    ctx.regen = True
    assert not _pdf_info.skip_file(ctx)
    assert 'forcing re-generation of all documents anyway' in caplog.text


def test_skip_file_document_updated(caplog):
    ctx = _get_should_skip_context(hash_='EDLM not the hash')
    assert not _pdf_info.skip_file(ctx)
    assert 'document updated, regenerating' in caplog.text


def test_skip_file_wrong_version(caplog):
    ctx = _get_should_skip_context(version='other version')
    assert not _pdf_info.skip_file(ctx)
    assert 'document generated with an older version of EDLM, regenerating' in caplog.text


def test_add_metadata():
    ctx = _pdf_info.Context()
    ctx.out_file = Path('./test')
    Path('./test').touch()
    writer = mock()
    when(writer).write()
    trailer = mock()
    trailer.Info = mock()
    when(pdfrw).PdfReader(str(ctx.out_file.absolute())).thenReturn(trailer)
    when(pdfrw).PdfFileWriter(str(ctx.out_file.absolute()), trailer=trailer).thenReturn(writer)
    when(_pdf_info)._get_document_hash(...).thenReturn('hash')
    _pdf_info.add_metadata_to_pdf(ctx)
    verify(writer).write()
    assert trailer.Info.Creator == 'EDLM ' + __version__
    assert trailer.Info.Producer == 'EDLM hash'



# coding=utf-8
"""
EDLM configuration
"""

import elib


class Config(elib.config.BaseConfig):
    """
    EDLM configuration
    """

    def __init__(self):
        elib.config.BaseConfig.__init__(self, __package__)

    miktex_archive = elib.config.ConfigProp(str, default='./miktex.7z')
    miktex_path = elib.config.ConfigProp(str, default='./miktex')
    pandoc_archive = elib.config.ConfigProp(str, default='./pandoc.7z')
    pandoc_path = elib.config.ConfigProp(str, default='./pandoc')
    debug = elib.config.ConfigProp(parser=bool, default=False)
    keep_temp_dir = elib.config.ConfigProp(parser=bool, default=False)


CFG = Config()



# coding=utf-8
"""
Gather the includes that are use throughout the document
"""

import typing
from pathlib import Path

import elib

from . import Context


def _gather_indices(folder: Path) -> typing.Iterator[Path]:
    for item in folder.iterdir():
        assert isinstance(item, Path)
        if item.is_file() and not item.name == 'index.md' and item.name.endswith('.md'):
            yield item.absolute()


def _gather_external_includes(ctx: Context) -> typing.Iterator[Path]:
    for item in ctx.source_folder.parent.iterdir():
        assert isinstance(item, Path)
        if item.is_dir() and Path(item, 'index.md').exists():
            yield item.absolute()


def _process_own_includes(ctx: Context):
    for include in _gather_indices(ctx.source_folder):
        relative_include = include.relative_to(ctx.source_folder)
        if f'//include "{relative_include}"' in ctx.markdown_text:
            ctx.includes.append(include)


def _process_external_includes(ctx: Context):
    for include in _gather_external_includes(ctx):
        relative_include = include.relative_to(ctx.source_folder.parent)
        if f'//include "{relative_include}"' in ctx.markdown_text:
            ctx.includes.append(include)


def get_includes(ctx: Context):
    """
    Gather the includes that are use throughout the document

    Args:
        ctx: Context
    """
    ctx.includes = []
    _process_own_includes(ctx)
    _process_external_includes(ctx)
    if ctx.includes:
        ctx.info(f'includes:\n{elib.pretty_format(ctx.includes)}')



# coding=utf-8

from edlm.convert import Context, Settings
from edlm.convert._preprocessor._markdown._references import Reference, process_references

DUMMY_REFS = {
    '//ref1': 'ref1_name, ref1_link',
    '//ref2': 'ref2_name, ref2_link',
    '//ref3': 'ref3_name, ref3_link',
}


def test_reference_class():
    name = 'name'
    link = 'link'
    raw_str = ','.join((name, link))
    abbrev = 'abbrev'
    ref = Reference(raw_str, abbrev)
    assert ref.abbrev == abbrev
    assert ref.name == name
    assert ref.link == link
    assert ref.to_latex() == f'\\href{{{link}}}{{{name}}}'
    assert ref.to_markdown() == f'[{name}]({link})'
    assert hash(ref) == hash(abbrev)


def test_process_references():
    ctx = Context()
    ctx.settings.references = DUMMY_REFS
    ctx.markdown_text = """
    This is some dummy markdown text.
    
    This should be ref1 //ref1.
    
    This should be ref3 //ref3.
    
    This should be ref1 //ref1 again.
    """
    process_references(ctx)
    assert ctx.latex_refs == ['\\href{ref1_link}{ref1_name}', '\\href{ref3_link}{ref3_name}']
    assert ctx.markdown_text == """
    This is some dummy markdown text.
    
    This should be ref1 [ref1_name](ref1_link).
    
    This should be ref3 [ref3_name](ref3_link).
    
    This should be ref1 [ref1_name](ref1_link) again.
    """


def test_process_references_no_ref():
    ctx = Context()
    ctx.settings = Settings()
    ctx.markdown_text = """
    This is some dummy markdown text.
    
    This should be ref1 //ref1.
    
    This should be ref3 //ref3.
    
    This should be ref1 //ref1 again.
    """
    process_references(ctx)
    assert ctx.latex_refs == []
    assert ctx.markdown_text == """
    This is some dummy markdown text.
    
    This should be ref1 //ref1.
    
    This should be ref3 //ref3.
    
    This should be ref1 //ref1 again.
    """



# coding=utf-8


from edlm.convert._preprocessor._markdown._aliases import Context, process_aliases


def test_process_aliases():
    ctx = Context()
    ctx.settings.data = {
        'aliases': {
            'key': 'value'
        }
    }
    ctx.markdown_text = 'this is some key text'
    process_aliases(ctx)
    assert ctx.markdown_text == 'this is some value text'



# coding=utf-8
"""
Command line interface
"""

import click
import elib

from edlm import LOGGER, __version__
from edlm.config import CFG
from edlm.convert import Context, make_pdf
from edlm.external_tools import MIKTEX, PANDOC


@click.group()
@click.version_option(version=__version__)
@click.option('--debug', default=False, help='Outputs DEBUG message on console', is_flag=True)
def cli(debug):
    """
    Command line interface
    """
    debug = debug or CFG.debug
    if debug:
        elib.custom_logging.set_handler_level('EDLM', 'ch', 'debug')
    else:
        elib.custom_logging.set_handler_level('EDLM', 'ch', 'info')

    LOGGER.info(__version__)

    PANDOC.setup()
    MIKTEX.setup()


@cli.group()
def convert():
    """
    Converts documents
    """


@convert.command()
@click.argument(
    'source_folder',
    type=click.Path(exists=True, file_okay=False, resolve_path=True, readable=True),
    nargs=-1,
)
@click.option('-k', '--keep-temp-dir', default=False, help='Keep temporary folder', is_flag=True)
@click.option('-f', '--force', default=False, help='Force re-generation of documents', is_flag=True)
def pdf(source_folder, keep_temp_dir, force):
    """
    Converts content of SOURCE_FOLDER(s) recursively for folders containing "index.md" files and convert them to PDF
    """

    ctx = Context()
    ctx.keep_temp_dir = keep_temp_dir or CFG.keep_temp_dir
    ctx.regen = force

    for folder in source_folder:
        make_pdf(ctx, folder)


# noinspection SpellCheckingInspection
if __name__ == '__main__':
    cli(obj={})  # pylint: disable=no-value-for-parameter,unexpected-keyword-arg
    exit(0)



# coding=utf-8
"""
Global context
"""
import copy
from pathlib import Path

import elib

from edlm import LOGGER

from ._settings import Settings

DEFAULT = {
    'source_folder': None,
    'paper_size': 'a4',
    'title': None,
    'markdown_text': None,
    'regen': False,

    'temp_dir': None,
    'keep_temp_dir': False,
    'media_folders': None,
    'index_file': None,
    'out_folder': None,
    'out_file': None,
    'source_file': None,

    'settings': None,
    'settings_files': None,

    'template_source': None,
    'template_file': None,

    'image_caption': None,
    'image_current': None,
    'image_extras': None,
    'image_max_width': None,
    'image_width': None,
    'image_width_str': None,
    'images_used': None,
    'images_unused': None,

    'latex_refs': None,

    'includes': None,
    'unprocessed_includes': None,
}


class _Val:
    def __init__(self, instance):
        self.instance = instance

    def _return_default(self):
        try:
            return DEFAULT[self.name]
        except KeyError:
            raise KeyError(f'CONTEXT: no default value for {self.name}')

    def __get__(self, instance, owner):
        try:
            return instance.data[self.name]
        except KeyError:
            return self._return_default()

    def __set__(self, instance, value):
        if value is not None and not isinstance(value, self.instance):
            raise TypeError(f'CONTEXT: expected "{self.instance}" for "{self.name}", got "{type(value)}"')
        instance.data[self.name] = value

    # this is the new initializer:
    def __set_name__(self, owner, name):
        assert owner
        self.name = name  # pylint: disable=attribute-defined-outside-init


class Context:
    """
    Global context
    """
    source_folder: Path = _Val(Path)
    paper_size: str = _Val(str)
    title: str = _Val(str)
    markdown_text: str = _Val(str)

    regen: bool = _Val(bool)  # flake8: noqa

    temp_dir: Path = _Val(Path)
    keep_temp_dir: bool = _Val(bool)
    media_folders: list = _Val(list)
    index_file: Path = _Val(Path)
    out_folder: Path = _Val(Path)
    out_file: Path = _Val(Path)
    source_file: Path = _Val(Path)

    settings: Settings = _Val(Settings)
    settings_files: list = _Val(list)

    template_source: Path = _Val(Path)
    template_file: Path = _Val(Path)

    image_caption: str = _Val(str)
    image_current: str = _Val(str)
    image_extras: str = _Val(str)
    image_max_width: int = _Val(int)
    image_width: int = _Val(int)
    image_width_str: str = _Val(str)
    images_used: set = _Val(set)
    images_unused: set = _Val(set)

    latex_refs: _Val(list)

    includes: list = _Val(list)
    unprocessed_includes: list = _Val(list)

    def __init__(self):
        self.data = {}
        self.skip_repr = ['markdown_text']
        self.settings = Settings()

    def get_sub_context(self):
        """
        Creates a copy of this context

        Returns: new context
        """
        new_context = Context()
        new_context.data = copy.deepcopy(self.data)
        return new_context

    def debug(self, text):
        """
        Convenient shortcut to main EDLM LOGGER
        """
        if self.index_file:
            LOGGER.debug(f'"{self.index_file}": {text}')
        else:
            LOGGER.debug(f'"{self.source_folder}": {text}')

    def info(self, text):
        """
        Convenient shortcut to main EDLM LOGGER
        """
        if self.index_file:
            LOGGER.info(f'"{self.index_file}": {text}')
        else:
            LOGGER.info(f'"{self.source_folder}": {text}')

    def error(self, text):
        """
        Convenient shortcut to main EDLM LOGGER
        """
        if self.index_file:
            LOGGER.error(f'"{self.index_file}": {text}')
        else:
            LOGGER.error(f'"{self.source_folder}": {text}')

    def warning(self, text):
        """
        Convenient shortcut to main EDLM LOGGER
        """
        if self.index_file:
            LOGGER.warning(f'"{self.index_file}": {text}')
        else:
            LOGGER.warning(f'"{self.source_folder}": {text}')

    def __repr__(self):  # pylint: disable=bad-continuation
        return elib.pretty_format({k: v for k, v in self.data.items() if k not in self.skip_repr})



# coding=utf-8
"""
Per-document settings
"""

import collections

import elib

DEFAULT = {
    'papersize': ['a4'],
    'aliases': dict(),
    'references': dict(),
}


class _Val:

    def __init__(self, instance):
        self.instance = instance

    def _return_default(self):
        try:
            return DEFAULT[self.name]
        except KeyError:
            raise KeyError(f'SETTINGS: no default value for {self.name}')

    def __get__(self, instance, owner):
        try:
            return instance.data[self.name]
        except KeyError:
            return self._return_default()

    def __set__(self, instance, value):
        if not isinstance(value, self.instance):
            raise TypeError(f'SETTINGS: expected "{self.instance}" for "{self.name}"')
        instance.data[self.name] = value

    # this is the new initializer:
    def __set_name__(self, owner, name):
        assert owner
        self.name = name  # pylint: disable=attribute-defined-outside-init


class Settings:
    """
    Per-document settings
    """
    papersize: list = _Val(list)
    aliases: dict = _Val(dict)
    references: dict = _Val(dict)

    def __init__(self):
        self.data = {}

    @staticmethod
    def _update_nested_dict(source_dict, updated_dict) -> collections.Mapping:
        """
        Updates a dictionary from another

        Args:
            source_dict: source dictionary (will be overwritten)
            updated_dict: updated dictionary (will take precedence)

        Returns: merged dictionary

        """
        for key, value in updated_dict.items():
            if isinstance(value, collections.Mapping):
                result = Settings._update_nested_dict(source_dict.get(key, {}), value)
                source_dict[key] = result
            else:
                source_dict[key] = updated_dict[key]
        return source_dict

    def update(self, other_dict: collections.Mapping):
        """
        Update these settings with another dictionary

        The other dictionary will have precedence

        Args:
            other_dict: dictionary to update from
        """
        self.data = self._update_nested_dict(self.data, other_dict)

    def __repr__(self) -> str:
        return "Settings:\n" + elib.pretty_format(self.data)



# coding=utf-8

from pathlib import Path

import elib
import pytest
from mockito import contains, patch, verify, when

from edlm.external_tools import MikTex, base


@pytest.fixture(name='mpm_file')
def _mpm_file():
    mpm_file = Path('./texmfs/config/miktex/config/miktex.ini')
    mpm_file.parent.mkdir(parents=True)
    yield mpm_file


@pytest.fixture(name='miktex')
def _miktex():
    when(base.BaseExternalTool)._is_installed().thenReturn(True)
    when(elib).run(contains('--version'), mute=True).thenReturn(('miktex version\ntext', 0))
    miktex = MikTex()
    miktex.install_dir = '.'
    yield miktex
    verify(base.BaseExternalTool)._is_installed()


def test_miktex_setup_create_new_mpm(miktex: MikTex):
    patch(MikTex, 'version', 'test')
    mpm_file = Path('./texmfs/config/miktex/config/miktex.ini')
    miktex.setup()
    assert mpm_file.exists()
    assert mpm_file.read_text(encoding='utf8') == """;;; DO NOT EDIT THIS FILE!


[MPM]
AutoInstall=1

"""


def test_miktex_setup_edit_mpm_add_section(mpm_file: Path, miktex: MikTex):
    mpm_file.write_text('some pre-existing content', encoding='utf8')
    patch(MikTex, 'version', 'test')
    miktex.setup()
    assert mpm_file.read_text(encoding='utf8') == """some pre-existing content
[MPM]
AutoInstall=1
"""


def test_miktex_setup_edit_mpm_add_line(mpm_file: Path, miktex: MikTex):
    mpm_file.write_text("""some pre-existing content
[MPM]
""", encoding='utf8')
    patch(MikTex, 'version', 'test')
    miktex.setup()
    assert mpm_file.read_text(encoding='utf8') == """some pre-existing content
[MPM]
AutoInstall=1
"""


def test_miktex_setup_edit_mpm_line(mpm_file: Path, miktex: MikTex):
    mpm_file.write_text(""";;; DO NOT EDIT THIS FILE!


[MPM]
AutoInstall=0

""", encoding='utf8')
    patch(MikTex, 'version', 'test')
    miktex.setup()
    assert mpm_file.read_text(encoding='utf8') == """;;; DO NOT EDIT THIS FILE!


[MPM]
AutoInstall=1

"""


def test_version(miktex: MikTex):
    when(base.BaseExternalTool).__call__('--version').thenReturn('miktex version\nnoise')
    miktex.setup()
    assert miktex.get_version() == 'version'



# coding=utf-8
from pathlib import Path

import pytest

import edlm.convert._get_settings as settings
from edlm.convert import Context, ConvertError, _settings


def test_update_nested_dict():
    source_dict = {
        'source': 'source',
        'shared': 'source',
        'sub': {
            'source': 'source',
            'shared': 'source',
        }
    }
    target_dict = {
        'target': 'target',
        'shared': 'target',
        'sub': {
            'target': 'target',
            'shared': 'target',
        }
    }
    settings.update_nested_dict(source_dict, target_dict)
    assert source_dict == {
        'source': 'source',
        'shared': 'target',
        'target': 'target',
        'sub': {
            'source': 'source',
            'shared': 'target',
            'target': 'target',
        }
    }


def test_update_nested_dict2():
    settings_ = _settings.Settings()
    source_dict = {
        'source': 'source',
        'shared': 'source',
        'sub': {
            'source': 'source',
            'shared': 'source',
        }
    }
    target_dict = {
        'target': 'target',
        'shared': 'target',
        'sub': {
            'target': 'target',
            'shared': 'target',
        }
    }
    settings_.update(source_dict)
    settings_.update(target_dict)
    assert settings_.data == {
        'source': 'source',
        'shared': 'target',
        'target': 'target',
        'sub': {
            'source': 'source',
            'shared': 'target',
            'target': 'target',
        }
    }


def test_no_setting_file():
    ctx = Context()
    ctx.source_folder = Path('.').absolute()
    with pytest.raises(ConvertError):
        settings.get_settings(ctx)


def test_single_settings():
    ctx = Context()
    ctx.source_folder = Path('.').absolute()
    settings_ = Path('./settings.yml')
    settings_.write_text('key: value')
    settings.get_settings(ctx)
    assert ctx.settings.data == {'key': 'value'}


def test_empty_settings():
    ctx = Context()
    ctx.source_folder = Path('.').absolute()
    settings_ = Path('./settings.yml')
    settings_.write_text('', encoding='utf8')
    with pytest.raises(ConvertError):
        settings.get_settings(ctx)


def test_nested_settings():
    sub1 = Path('./sub1').absolute()
    sub1.mkdir()
    settings_1 = Path(sub1, 'settings.yml').absolute()
    settings_1.write_text(
        'key1: value1\n'
        'key2: value1\n'
        'key3: value1\n'
        'key4:\n'
        '  key1: value1\n'
        '  key2: value1\n'
        '  key3: value1\n',
        encoding='utf8'
    )
    sub2 = Path(sub1, 'sub2').absolute()
    sub2.mkdir()
    settings_2 = Path(sub2, 'settings.yml').absolute()
    settings_2.write_text(
        'key2: value2\n'
        'key4:\n'
        '  key2: value2\n',
        encoding='utf8'
    )
    sub3 = Path(sub2, 'sub3').absolute()
    sub3.mkdir()
    settings_3 = Path(sub3, 'settings.yml').absolute()
    settings_3.write_text(
        'key3: value3\n'
        'key4:\n'
        '  key3: value3\n',
        encoding='utf8'
    )
    ctx = Context()
    ctx.source_folder = sub3
    settings.get_settings(ctx)
    assert ctx.settings.data == {
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'key4': {
            'key1': 'value1',
            'key2': 'value2',
            'key3': 'value3',
        },
    }


def test_no_default():
    class Dummy(_settings.Settings):
        no_default = _settings._Val(str)

    settings_ = Dummy()
    with pytest.raises(KeyError):
        _ = settings_.no_default


def test_wrong_type():
    settings_ = _settings.Settings()
    with pytest.raises(TypeError):
        settings_.references = 'test'



# coding=utf-8
"""
Processes Tex templates
"""

import elib
from jinja2 import BaseLoader, Environment, TemplateNotFound

from .._context import Context
from .._exc import ConvertError


class TexTemplateLoader(BaseLoader):
    """
    Loads the template for Jinja2
    """

    def __init__(self, ctx: Context):
        self.ctx = ctx

    def get_source(self, environment, template):
        # noinspection SpellCheckingInspection
        """
        Get the template source, filename and reload helper for a template.
        It's passed the environment and template name and has to return a
        tuple in the form ``(source, filename, uptodate)`` or raise a
        `TemplateNotFound` error if it can't locate the template.

        The source part of the returned tuple must be the source of the
        template as unicode string or a ASCII bytestring.  The filename should
        be the name of the file on the filesystem if it was loaded from there,
        otherwise `None`.  The filename is used by python for the tracebacks
        if no loader extension is used.

        The last item in the tuple is the `uptodate` function.  If auto
        reloading is enabled it's always called to check if the template
        changed.  No arguments are passed so the function must store the
        old state somewhere (for example in a closure).  If it returns `False`
        the template will be reloaded.
        """
        self.ctx.debug(f'environment: {elib.pretty_format(environment)}')
        self.ctx.debug(f'template: {template}')
        if not self.ctx.template_source.exists():
            raise TemplateNotFound(str(self.ctx.template_source))
        source = self.ctx.template_source.read_text(encoding='utf8')
        # "False" here means always reload the template
        return source, self.ctx.template_source.absolute(), False


def _get_jinja_env(ctx: Context):
    return Environment(
        block_start_string=r'\BLOCK{',
        block_end_string=r'}',
        variable_start_string=r'\VAR{',
        variable_end_string=r'}',
        comment_start_string=r'\#{',
        comment_end_string=r'}',
        line_statement_prefix=r'%%',
        line_comment_prefix=r'%#',
        trim_blocks=True,
        autoescape=False,
        loader=TexTemplateLoader(ctx)
    )


def process_tex_template(ctx: Context):
    """
    Processes the Tex template

    Adds path the media folders
    """
    ctx.debug(f'processing Tex template file: {ctx.template_source}')

    try:
        elib.path.ensure_file(ctx.template_source)
    except FileNotFoundError:
        raise ConvertError(f'LaTeX template not found: "{ctx.template_source}"', ctx)

    jinja_env = _get_jinja_env(ctx)

    media_folders = [folder.absolute() for folder in ctx.media_folders]
    media_folders = [str(folder).replace('\\', '/') for folder in media_folders]
    media_folders = ''.join(f'{{{folder}/}}' for folder in media_folders)
    ctx.debug(f'adding media folders to template: {media_folders}')
    try:
        template = jinja_env.get_template(ctx.template_source.name)
        ctx.template_file.write_text(template.render(**locals()), encoding='utf8')
    except TemplateNotFound:
        raise ConvertError('LaTeX template not found', ctx)



# coding=utf-8
"""
Gathers the media folders
"""

import elib

from ._context import Context


def get_media_folders(ctx: Context):
    """
    Gathers the media folders
    """
    ctx.info('gathering media folders')

    media_folders = []

    this_folder = ctx.source_folder
    while True:
        ctx.debug(f'traversing: "{this_folder}"')
        media_folder_candidate = elib.path.ensure_path(this_folder, 'media', must_exist=False).absolute()
        if media_folder_candidate.exists() and media_folder_candidate.is_dir():
            ctx.debug(f'media folder found: "{media_folder_candidate}"')
            media_folders.append(media_folder_candidate)
        if len(this_folder.parents) is 1:
            ctx.debug(f'reach mount point at: "{this_folder}"')
            break
        this_folder = this_folder.parent

    # if not media_folders:
    #     raise ConvertError('no media folder found', ctx)

    ctx.info(f'media folders:\n{elib.pretty_format(media_folders)}')
    ctx.media_folders = media_folders



# coding=utf-8
"""
Miktex external tool
"""

import typing
from pathlib import Path

import elib

from edlm import HERE
from edlm.external_tools.base import BaseExternalTool

STR_OR_PATH = typing.Union[str, Path]

MPM_CONFIG = """;;; DO NOT EDIT THIS FILE!


[MPM]
AutoInstall=1

"""


class MikTex(BaseExternalTool):
    """
    Miktex external tool
    """
    url = r'https://www.dropbox.com/s/ivyb6s5itb5len2/miktex.7z?dl=1'
    # noinspection SpellCheckingInspection
    hash = 'fc20affd161264e7c5b816ddc85955cd'
    default_archive = Path(HERE, 'miktex.7z').absolute()
    default_install = Path(HERE, 'miktex').absolute()
    expected_version = '2.9.6354'

    def get_version(self) -> str:
        """
        Returns: Miktex version
        """
        if self._version is None:
            self._version = self('--version').split('\n')[0].split(' ')[1]
        return self._version

    def get_exe(self) -> Path:
        """
        Returns: Miktex executable
        """
        if self._exe is None:
            self._exe = Path(self.install_dir, 'miktex/texmfs/install/miktex/bin/pdflatex.exe').absolute()
        return self._exe

    @staticmethod
    def _create_new_mpm_settings_file(mpm_config_file):
        elib.path.ensure_dir(mpm_config_file.parent, must_exist=False, create=True)
        mpm_config_file.write_text(MPM_CONFIG, encoding='utf8')

    @staticmethod
    def _edit_auto_install_line(content):
        for index, line in enumerate(content):
            if 'AutoInstall=' in line:
                content[index] = 'AutoInstall=1'
                return True
        return False

    @staticmethod
    def _add_auto_install_line(content: list):
        for index, line in enumerate(content):
            if '[MPM]' in line:
                content.insert(index + 1, 'AutoInstall=1')
                return True
        return False

    @staticmethod
    def _add_mpm_section(content: list):
        content.append('[MPM]')
        content.append('AutoInstall=1\n')
        return True

    def _edit_existing_mpm_settings_file(self, mpm_config_file):
        content = mpm_config_file.read_text(encoding='utf8').split('\n')
        func_list = [self._edit_auto_install_line, self._add_auto_install_line, self._add_mpm_section]
        for func in func_list:  # pragma: no cover
            if func(content):
                mpm_config_file.write_text('\n'.join(content))
                return

    def _write_mpm_settings_file(self):
        mpm_config_file = Path(self.install_dir, 'texmfs/config/miktex/config/miktex.ini')
        if not mpm_config_file.exists():
            self._create_new_mpm_settings_file(mpm_config_file)
        else:
            self._edit_existing_mpm_settings_file(mpm_config_file)

    def setup(self):
        """
        Setup Miktex
        """
        super(MikTex, self).setup()
        self._write_mpm_settings_file()



# coding=utf-8

from pathlib import Path

import pytest

from edlm.convert import _get_index


def test_get_index():
    ctx = _get_index.Context()
    ctx.source_folder = Path('.').absolute()
    with pytest.raises(FileNotFoundError):
        _get_index.get_index_file(ctx)
    Path('./index.md').mkdir()
    with pytest.raises(TypeError):
        _get_index.get_index_file(ctx)
    Path('./index.md').rmdir()
    Path('./index.md').touch()
    _get_index.get_index_file(ctx)
    assert ctx.index_file == Path('./index.md').absolute()



# coding=utf-8
"""
Replaces include directive with the actual markdown content
"""

from pathlib import Path

import elib

from . import Context


class Inclusion:
    """
    Dummy context for a file inclusion
    """
    sub_context: Context
    include_str: str
    markdown_source: Path
    parent_folder: Path


def _process_include(ctx: Context, inclusion: Inclusion):
    from . import process_markdown
    from ..._get_includes import get_includes
    from ..._get_media_folders import get_media_folders
    inclusion.sub_context.markdown_text = inclusion.markdown_source.read_text(encoding='utf8')
    inclusion.sub_context.index_file = inclusion.markdown_source
    inclusion.sub_context.includes = []
    get_includes(inclusion.sub_context)
    get_media_folders(inclusion.sub_context)
    process_markdown(inclusion.sub_context)
    ctx.markdown_text = ctx.markdown_text.replace(inclusion.include_str, inclusion.sub_context.markdown_text)
    if inclusion.sub_context.latex_refs:
        refs = set(inclusion.sub_context.latex_refs)
        refs.update(ctx.latex_refs)
        ctx.latex_refs = list(refs)


def _process_local_include(ctx: Context, inclusion: Inclusion):
    inclusion.include_str = f'//include "{inclusion.parent_folder.relative_to(ctx.source_folder)}"'
    inclusion.markdown_source = Path(ctx.source_folder, inclusion.parent_folder)
    _process_include(ctx, inclusion)
    if inclusion.sub_context.images_used:
        ctx.images_used.update(inclusion.sub_context.images_used)


def _process_external_include(ctx: Context, inclusion: Inclusion):
    inclusion.include_str = f'//include "{inclusion.parent_folder.relative_to(ctx.source_folder.parent)}"'
    inclusion.markdown_source = Path(inclusion.parent_folder, 'index.md')
    inclusion.sub_context.source_folder = inclusion.parent_folder
    _process_include(ctx, inclusion)


def _get_unprocessed_includes(ctx: Context):
    for line_nr, line in enumerate(ctx.markdown_text.split('\n')):
        if '//include' in line:
            ctx.unprocessed_includes.append(f'line {line_nr+1:05d}: {line}')


def process_includes(ctx: Context):
    """
    Replaces include directive with the actual markdown content

    Args:
        ctx: Context
    """
    ctx.unprocessed_includes = []
    for include in ctx.includes:
        inclusion = Inclusion()
        inclusion.sub_context = ctx.get_sub_context()
        inclusion.parent_folder = include
        if include.is_file():
            _process_local_include(ctx, inclusion)
        elif include.is_dir():
            _process_external_include(ctx, inclusion)

    _get_unprocessed_includes(ctx)
    if ctx.unprocessed_includes:
        ctx.warning(f'there are unprocessed "//include" directives:\n{elib.pretty_format(ctx.unprocessed_includes)}')



# coding=utf-8


from pathlib import Path

from mockito import verifyStubbedInvocationsAreUsed, when

from edlm.convert import Context
from edlm.convert._preprocessor import _markdown


def test_markdown_preprocessor():
    ctx = Context()
    index_file = Path('./index.md')
    index_file.write_text('', encoding='utf8')
    ctx.index_file = index_file
    ctx.includes = []
    ctx.markdown_text = ''
    when(_markdown).process_aliases(ctx)
    when(_markdown).process_images(ctx)
    when(_markdown).process_references(ctx)
    _markdown.process_markdown(ctx)
    verifyStubbedInvocationsAreUsed()



# coding=utf-8
"""
Get templates folder
"""

from pathlib import Path

from . import Context, ConvertError


def get_template(ctx: Context):
    """
    Get templates folder
    """

    ctx.info('looking for template')

    ctx.template_source = None

    this_folder = ctx.source_folder
    while not ctx.template_source:
        ctx.debug(f'traversing: {this_folder}')
        file = Path(this_folder, 'template.tex').absolute()
        if file.exists() and file.is_file():
            ctx.template_source = file
        if len(this_folder.parents) is 1:
            raise ConvertError('no "template.tex" file found', ctx)
        this_folder = this_folder.parent

    ctx.info(f'using template file: {ctx.template_source}')



# coding=utf-8
"""
Processes references in the Markdown text
"""

from . import Context

REFS_TEMPLATE = r"""
\newpage
# References
{}
"""


class Reference:
    """
    Represents a reference in the document
    """

    def __init__(self, raw_ref, abbrev):
        self._abbrev = abbrev
        self._raw_ref = raw_ref
        self._name = raw_ref.split(',')[0].lstrip().rstrip()
        self._link = raw_ref.split(',')[1].lstrip().rstrip()

    @property
    def abbrev(self):
        """
        Returns: reference's abbreviation
        """
        return self._abbrev

    @property
    def name(self):
        """
        Returns: reference's name
        """
        return self._name

    @property
    def link(self):
        """
        Returns: reference's link
        """
        return self._link

    def to_latex(self):
        """
        Returns: convert reference to Latex text
        """
        return f'\\href{{{self.link}}}{{{self.name}}}'

    def to_markdown(self):
        """
        Returns: convert reference to Markdown text
        """
        return f'[{self.name}]({self.link})'

    def __hash__(self):
        return hash(self.abbrev)

    def __repr__(self):
        return self.name


def process_references(ctx: Context):
    """
    Processes references in the Markdown text

    Args:
        ctx: Context
    """
    ctx.latex_refs = list()
    for abbrev, raw_ref in ctx.settings.references.items():
        if abbrev in ctx.markdown_text:
            ref = Reference(raw_ref, abbrev)
            ctx.latex_refs.append(ref.to_latex())
            ctx.markdown_text = ctx.markdown_text.replace(abbrev, ref.to_markdown())
            ctx.debug(f'used reference: {abbrev}')
    ctx.latex_refs = sorted(ctx.latex_refs)



# coding=utf-8
"""
Pytest config
"""

import os
import sys

import pytest
from click.testing import CliRunner
from mockito import unstub


@pytest.fixture(scope='session')
def miktex_path():
    yield os.path.abspath(
        os.path.join(
            os.path.dirname(
                os.path.dirname(__file__)
            ),
            'miktex',
        )
    )


@pytest.fixture()
def runner():
    runner_ = CliRunner()
    yield runner_


# noinspection PyUnusedLocal
def pytest_configure(config):
    """Setup"""
    assert config
    sys.called_from_test = True
    sys.path.append('.')


def pytest_unconfigure(config):
    """Tear down"""
    # noinspection PyUnresolvedReferences
    del sys.called_from_test
    assert config
    sys.path.remove('.')


@pytest.fixture(autouse=True)
def cleandir(request, tmpdir):
    """Provides a clean working dir"""
    if 'nocleandir' in request.keywords:
        yield
    else:
        current_dir = os.getcwd()
        os.chdir(str(tmpdir))
        yield os.getcwd()
        os.chdir(current_dir)


@pytest.fixture(autouse=True)
def _unstub():
    unstub()
    yield
    unstub()


@pytest.fixture(autouse=True)
def _reset_cache():
    from edlm.external_tools.base import _find_patool
    _find_patool.cache_clear()
    yield
    _find_patool.cache_clear()


@pytest.fixture(autouse=True)
def _setup_config():
    from edlm.config import CFG
    CFG.keep_temp_dir = False
    CFG.debug = False
    yield


def pytest_addoption(parser):
    """Add option for long tests"""
    parser.addoption("--long", action="store_true",
                     help="run long tests")


def pytest_runtest_setup(item):
    """Skip long tests"""
    long_marker = item.get_marker("long")
    if long_marker is not None and not item.config.getoption('long'):
        pytest.skip('skipping long tests')



# coding=utf-8
"""
EDLM setup file
"""

from pip.req import parse_requirements
from setuptools import find_packages, setup

requirements = [str(r.req) for r in
                parse_requirements('requirements.txt', session=False)]
test_requirements = [str(r.req) for r in
                     parse_requirements('requirements-dev.txt', session=False)]

CLASSIFIERS = filter(None, map(str.strip,
                               """
Development Status :: 2 - Pre-Alpha
Environment :: Console
Environment :: Win32 (MS Windows)
Intended Audience :: End Users/Desktop
Natural Language :: English
Operating System :: Microsoft :: Windows :: Windows 7
Operating System :: Microsoft :: Windows :: Windows 8
Operating System :: Microsoft :: Windows :: Windows 8.1
Operating System :: Microsoft :: Windows :: Windows 10
License :: OSI Approved :: MIT License
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Topic :: Documentation
Topic :: Text Processing
Topic :: Text Processing :: Markup
Topic :: Text Processing :: Markup :: LaTeX
Topic :: Utilities
""".splitlines()))

entry_points = '''
[console_scripts]
edlm=edlm.cli:cli
'''

setup(
    name='edlm',
    package_dir={'edlm': 'edlm'},
    use_scm_version=True,
    install_requires=requirements,
    tests_require=test_requirements,
    entry_points=entry_points,
    setup_requires=['setuptools_scm'],
    test_suite='pytest',
    packages=find_packages('.'),
    python_requires='>=3.6',
    license='MIT',
    classifiers=CLASSIFIERS,
)



# coding=utf-8
"""
Convert package exceptions
"""

from ._context import Context


class ConvertError(Exception):
    """
    Raised when an error happen during the conversion
    """

    def __init__(self, message, ctx: Context):
        super(ConvertError, self).__init__(f'"{ctx.source_folder}": conversion error: {message}')
        self.ctx = ctx



# coding=utf-8
"""
Manages temporary folder
"""

import shutil
import tempfile

import elib

from edlm import LOGGER

from ._context import Context


class TempDir:
    """
    Manages temporary folder
    """

    def __init__(self, ctx: Context):
        self.ctx = ctx
        self.path = elib.path.ensure_dir(tempfile.mkdtemp(dir='.', prefix='__TMP')).absolute()

    def __enter__(self):
        LOGGER.debug(f'"{self.ctx.source_folder}": using temporary folder: "{self.path}"')
        self.ctx.temp_dir = self.path

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type:
            if self.ctx.keep_temp_dir:
                LOGGER.info(f'"{self.ctx.source_folder}": build successful, keeping temp dir: "{self.path}"')
            else:
                LOGGER.debug(f'"{self.ctx.source_folder}": build successful, removing temp dir')
                shutil.rmtree(self.path)
                self.ctx.temp_dir = None
        else:
            LOGGER.warning(f'"{self.ctx.source_folder}": build failed, keeping temp dir: "{self.path}"')



# coding=utf-8

import pytest

from edlm.convert._temp_folder import Context, TempDir


def test_temp_dir():
    ctx = Context()
    assert ctx.temp_dir is None
    with TempDir(ctx):
        assert ctx.temp_dir.is_dir()
        assert ctx.temp_dir.exists()
        temp_dir = ctx.temp_dir

    assert ctx.temp_dir is None
    assert not temp_dir.exists()


def test_temp_dir_keep():
    ctx = Context()
    ctx.keep_temp_dir = True
    assert ctx.temp_dir is None
    with TempDir(ctx):
        assert ctx.temp_dir.is_dir()
        assert ctx.temp_dir.exists()
        temp_dir = ctx.temp_dir

    assert ctx.temp_dir == temp_dir
    assert temp_dir.exists()


def test_temp_dir_with_error():
    ctx = Context()
    assert ctx.temp_dir is None
    with pytest.raises(TypeError):
        with TempDir(ctx):
            assert ctx.temp_dir.is_dir()
            assert ctx.temp_dir.exists()
            temp_dir = ctx.temp_dir
            raise TypeError()

    assert ctx.temp_dir == temp_dir
    assert temp_dir.exists()



# coding=utf-8

from pathlib import Path

import pytest
from jinja2 import Environment, TemplateNotFound
from mockito import when

from edlm.convert import Context, ConvertError
from edlm.convert._preprocessor import _tex_template as tex


def test_template_loader():
    ctx = Context()
    template = Path('./template.tex').absolute()
    template.write_text('moo', encoding='utf8')
    ctx.template_source = template
    env = Environment(autoescape=True)
    loader = tex.TexTemplateLoader(ctx)
    source, template_, reload = loader.get_source(env, 'template.tex')
    assert source == 'moo'
    assert template_ == template
    assert reload is False


def test_template_loader_no_template():
    ctx = Context()
    template = Path('./template.tex').absolute()
    ctx.template_source = template
    env = Environment(autoescape=True)
    loader = tex.TexTemplateLoader(ctx)
    with pytest.raises(TemplateNotFound):
        loader.get_source(env, 'template.tex')


def test_process_template_no_template():
    ctx = Context()
    template = Path('./template.tex').absolute()
    ctx.template_source = template
    with pytest.raises(ConvertError):
        tex.process_tex_template(ctx)


def test_process_template_latex_error():
    ctx = Context()
    template = Path('./template.tex').absolute()
    ctx.template_source = template
    ctx.media_folders = []
    template.touch()
    when(tex.TexTemplateLoader).get_source(...).thenRaise(TemplateNotFound(ctx))
    ctx.template_source = template
    with pytest.raises(ConvertError):
        tex.process_tex_template(ctx)


def test_process_template():
    ctx = Context()
    template = Path('./template.tex').absolute()
    template_out = Path('./out.tex').absolute()
    ctx.template_file = template_out
    ctx.template_source = template
    ctx.media_folders = []
    template.touch()
    ctx.template_source = template
    tex.process_tex_template(ctx)
    assert ctx.template_file.read_text(encoding='utf8') == ''
